"""
Borg Legacy Package — Borg 1.x compatibility layer.

This package contains all code specific to reading/writing Borg 1.x repositories.
It is required for ``borg transfer --from-borg1`` and ``borg serve`` (serving v1 clients).

This package can be removed entirely when Borg 1.x support is dropped.
"""
